from datetime import datetime

from elasticsearch.client import Elasticsearch

from jinja2 import Environment, PackageLoader, select_autoescape

from pyramid_debugtoolbar.panels import DebugPanel


env = Environment(
    loader=PackageLoader("cubicweb_elasticsearch.pviews", "templates"),
    autoescape=select_autoescape(["html", "xml"]),
)


class ESDebugPanel(DebugPanel):
    """ES queries panel"""

    name = "ES"
    has_content = True
    template = "cubicweb_elasticsearch.pviews:templates/espanel.jinja2"
    title = "Queries"
    nav_title = "ElasticSearch"

    def __init__(self, request):
        self.data = {"request_path": request.path_info, "queries": {}}
        self.queries = []
        orig_search = Elasticsearch.search

        def traced_search(es_client, index=None, doc_type=None, body=None, **params):
            start = datetime.now()
            query = {
                "index": index,
                "doc_type": doc_type,
                "body": body,
                "params": params,
                "starttime": start,
                "response": {},
                "elapsed": 0,
            }
            self.queries.append(query)
            result = orig_search(es_client, index, doc_type, body, **params)
            elapsed = datetime.now() - start
            miliseconds = elapsed.seconds * 1000 + elapsed.microseconds / 1000
            query.update(
                {
                    "response": {
                        "hits": result.get("hits", []),
                        "aggregations": result.get("aggregations", {}),
                    },
                    "elapsed": miliseconds,
                }
            )
            return result

        Elasticsearch.search = traced_search

    def render_content(self, request):
        template = env.get_template("espanel.jinja2")
        return template.render(queries=self.queries)


def includeme(config):
    config.add_debugtoolbar_panel(ESDebugPanel)



# copyright 2016-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import unittest
from elasticsearch_dsl import Search

from cubicweb.devtools import testlib
from cubicweb_elasticsearch.search_helpers import compose_search
from logilab.mtconverter import xml_escape


class ComposeSearchTestCase(testlib.TestCase):
    def compose_search(self, query):
        self.search = Search()
        return compose_search(self.search, query)

    def test_simple(self):
        self.assertTrue(self.compose_search("test").to_dict())

    def test_two_terms(self):
        self.assertTrue(self.compose_search("test this").to_dict())

    def test_one_quote(self):
        # only one quote is not a phrase type search
        self.assertTrue(self.compose_search("test 'this").to_dict())

    def test_two_with_quotes(self):
        self.assertTrue(self.compose_search('"test this"').to_dict())
        self.assertTrue(self.compose_search("'test this'").to_dict())
        self.assertTrue(self.compose_search(xml_escape("'test this'")).to_dict())

    # def test_three_with_quotes(self):
    #     self.assertTrue(self.compose_search('"test this" this_too').to_dict(),
    #                       Q('bool',
    #                         must=[Q('multi_match', query='test', fields=()).to_dict(),
    #                               Q('multi_match', query='this', fields=())],
    #                         should=).to_dict())

    def test_two_with_negate(self):
        self.assertTrue(self.compose_search("test -this").to_dict())

    def test_two_with_or(self):
        self.assertTrue(self.compose_search("test or this").to_dict())
        # self.assertTrue(self.compose_search('test ou this'),
        #                   self.search.query('bool',
        #                     should=[Q('multi_match', query='test', fields=()),
        # Q('multi_match', query='this', fields=())]).to_dict())

    def test_or_on_its_own(self):
        self.assertTrue(self.compose_search("or").to_dict())


if __name__ == "__main__":
    unittest.main()



import unittest

from mock import patch

from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_elasticsearch.es import indexable_entities


class UtilTestCase(CubicWebTC):
    @patch("cubicweb.server.session.Connection.drop_entity_cache")
    def test_indexable_entities(self, drop):
        """check indexable_entities yield entities as expected"""
        with self.admin_access.cnx() as cnx:
            print(type(cnx))
            p1 = cnx.create_entity("Person", name=u"p1")
            p2 = cnx.create_entity("Person", name=u"p3")
            p3 = cnx.create_entity("Person", name=u"p3")
            drop.reset_mock()
            self.assertEqual(
                [p.eid for p in indexable_entities(cnx, "Person")],
                [p1.eid, p2.eid, p3.eid],
            )
            self.assertEqual(drop.call_count, 1)
            drop.reset_mock()
            # now try again with chunksize=2 to make sure we have split
            # rset in 2
            self.assertEqual(
                [p.eid for p in indexable_entities(cnx, "Person", chunksize=2)],
                [p1.eid, p2.eid, p3.eid],
            )
            self.assertEqual(drop.call_count, 2)


if __name__ == "__main__":
    unittest.main()



from __future__ import print_function

import sys
import unittest
from io import StringIO

from mock import patch

from elasticsearch_dsl.faceted_search import FacetedResponse
from elasticsearch_dsl.search import Search

from cubicweb.devtools import testlib
from cubicweb.cwconfig import CubicWebConfiguration
from cubicweb_elasticsearch import ccplugin
from cubicweb_elasticsearch.es import indexable_types, fulltext_indexable_rql


# TODO - find a way to configure ElasticSearch as non threaded while running tests
# so that the traces show the full stack, not just starting from connection.http_*
class ExportElasticSearchTC(testlib.AutoPopulateTest):
    # ignore ComputedRelations
    ignored_relations = set(
        (
            "narrower_concept",
            "hidden_label",
            "preferred_label",
            "alternative_label",
        )
    )

    def setup_database(self):
        super(ExportElasticSearchTC, self).setup_database()
        self.orig_config_for = CubicWebConfiguration.config_for
        config_for = lambda appid, *args, **kwargs: self.config  # noqa
        CubicWebConfiguration.config_for = staticmethod(config_for)
        self.config[
            "elasticsearch-locations"
        ] = "http://nonexistant.elastic.search:9200"
        self.config["index-name"] = "unittest_index_name"
        self.tested_etype = "Person"

    def to_test_etypes(self):
        with self.admin_access.repo_cnx() as cnx:
            types = indexable_types(cnx.repo)
        return types

    def tearDown(self):
        CubicWebConfiguration.config_for = self.orig_config_for
        super(ExportElasticSearchTC, self).tearDown()

    def test_indexable_types(self):
        with self.admin_access.repo_cnx() as cnx:
            self.assertIn(self.tested_etype, self.to_test_etypes())
            self.assertIn(self.tested_etype, indexable_types(cnx.vreg.schema))

    @patch("elasticsearch.client.Elasticsearch.index", unsafe=True)
    @patch("elasticsearch.client.Elasticsearch.bulk", unsafe=True)
    @patch("elasticsearch.client.indices.IndicesClient.exists", unsafe=True)
    @patch("elasticsearch.client.indices.IndicesClient.create", unsafe=True)
    def test_ccplugin(self, create, exists, bulk, index):
        # TODO disable hook!!! then remove index mock
        with self.admin_access.repo_cnx() as cnx:
            with cnx.allow_all_hooks_but("es"):
                self.auto_populate(10)
        bulk.reset_mock()
        cmd = [self.appid, "--dry-run"]
        sys.stdout = out = StringIO()
        try:
            ccplugin.IndexInES(None).main_run(cmd)
        finally:
            sys.stdout = sys.__stdout__
        self.assertEqual("", out.getvalue())
        create.assert_not_called()
        bulk.assert_not_called()

        # TODO try wrong option
        # cmd = [self.appid, '--wrong-option', 'yes']

        cmd = [self.appid]
        sys.stdout = StringIO()
        try:
            ccplugin.IndexInES(None).main_run(cmd)
        finally:
            sys.stdout = sys.__stdout__
        with self.admin_access.repo_cnx() as cnx:
            self.assertTrue(
                cnx.execute("Any X WHERE X is %(etype)s" % {"etype": self.tested_etype})
            )
        # TODO - put this somewhere where it tests on the first get_connection
        # create.assert_called_with(ignore=400,
        #                          index='unittest_index_name',
        #                          body=INDEX_SETTINGS)
        bulk.assert_called()
        # TODO ? check called data

    @patch("elasticsearch.client.indices.IndicesClient.create", unsafe=True)
    @patch("elasticsearch.client.indices.IndicesClient.exists", unsafe=True)
    @patch("elasticsearch.client.Elasticsearch.index", unsafe=True)
    def test_es_hooks_create(self, index, exists, create):
        with self.admin_access.cnx() as cnx:
            cnx.create_entity(self.tested_etype, name=u"Jean Valjean", age=14)
            cnx.commit()
            index.assert_called()

    @patch("elasticsearch.client.indices.IndicesClient.create", unsafe=True)
    @patch("elasticsearch.client.indices.IndicesClient.exists", unsafe=True)
    @patch("elasticsearch.client.Elasticsearch.index", unsafe=True)
    def test_es_hooks_modify(self, index, exists, create):
        with self.admin_access.cnx() as cnx:
            entity = cnx.create_entity(self.tested_etype, name=u"Jean Valjean")
            cnx.commit()
            index.reset_mock()
            entity.cw_set(name=u"Victor Hugo")
            cnx.commit()
            index.assert_called()


def mock_execute_150(*args, **kwargs):
    return mock_execute(100)


def mock_execute_15(*args, **kwargs):
    return mock_execute(15)


def mock_execute_1(*args, **kwargs):
    return mock_execute(1)


def mock_execute(count):
    def _result(i):
        return {
            "_source": {
                "cwuri": "http://example.org/{}".format(i),
                "eid": i,
                "cw_etype": "Person",
                "name": "Jean Valjean",
            },
            "_type": "Person",
            "_score": 1,
        }

    search = Search(doc_type="_doc", index="unittest_index_name")
    return FacetedResponse(
        search,
        {
            "hits": {
                "hits": [_result(i) for i in range(count)],
                "total": {"value": count, "relation": "eq"},
            }
        },
    )


def mock_cnx(*args, **kwargs):
    return True


class ElasticSearchViewsTC(testlib.CubicWebTC):

    # TODO generate X tests ranging the number of results from 1 to 150
    @patch("elasticsearch_dsl.search.Search.execute", new=mock_execute_1)
    @patch("elasticsearch_dsl.connections.connections.get_connection", new=mock_cnx)
    def test_search_view_1(self):
        with self.new_access("anon").web_request() as req:
            # self._cw.form.get('search'))
            self.view("esearch", req=req, template=None)

    @patch("elasticsearch_dsl.search.Search.execute", new=mock_execute_15)
    @patch("elasticsearch_dsl.connections.connections.get_connection", new=mock_cnx)
    def test_search_view_15(self):
        with self.new_access("anon").web_request() as req:
            # self._cw.form.get('search'))
            self.view("esearch", req=req, template=None)

    @patch("elasticsearch_dsl.search.Search.execute", new=mock_execute_150)
    @patch("elasticsearch_dsl.connections.connections.get_connection", new=mock_cnx)
    def skip_test_search_view_150(self):
        with self.new_access("anon").web_request() as req:
            # self._cw.form.get('search'))
            self.view("esearch", req=req, template=None)


class ElasticsearchTC(testlib.CubicWebTC):
    def test_1(self):
        with self.admin_access.cnx() as cnx:
            etype = "Person"
            rql = fulltext_indexable_rql(etype, cnx)
            self.assertIn("age", rql)
            self.assertNotIn("eid", rql)
            self.assertEqual(rql.count("modification_date"), 1)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# copyright 2016-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-elasticsearch entity's classes"""

import collections

from functools import partial

from urllib3.exceptions import ProtocolError

from elasticsearch.exceptions import ConnectionError, NotFoundError

from logilab.common.decorators import cachedproperty

from cubicweb import view, neg_role
from cubicweb.predicates import is_instance

from cubicweb.appobject import AppObject

from cubicweb_elasticsearch import es


def deep_update(d1, d2):
    for key, value in d2.iteritems():
        if isinstance(value, collections.Mapping):
            d1[key] = deep_update(d1.get(key, {}), value)
        else:
            d1[key] = d2[key]
    return d1


class EsRegistry(AppObject):
    __registry__ = "es"


class Indexer(EsRegistry):
    __regid__ = "indexer"
    adapter = "IFullTextIndexSerializable"
    settings = {
        "settings": {
            "analysis": {
                "analyzer": {
                    "default": {
                        "filter": ["my_ascii_folding", "lowercase", "french_snowball"],
                        "tokenizer": "standard",
                    }
                },
                "filter": {
                    "my_ascii_folding": {
                        "preserve_original": True,
                        "type": "asciifolding",
                    },
                    "french_snowball": {"type": "snowball", "language": "French"},
                },
            },
        }
    }

    @property
    def index_name(self):
        return self._cw.vreg.config["index-name"]

    def get_connection(self):
        self.create_index()
        return es.get_connection(self._cw.vreg.config)

    def create_index(self, index_name=None, custom_settings=None):
        index_name = index_name or self.index_name
        if custom_settings is None:
            settings = self.settings
        else:
            settings = {}
            deep_update(settings, self.settings)
            deep_update(settings, custom_settings)
        es_cnx = es.get_connection(self._cw.vreg.config)
        if es_cnx is not None:
            es.create_index(es_cnx, index_name, settings)

    def es_index(self, entity, params=None):
        es_cnx = self.get_connection()
        if es_cnx is None or not self.index_name:
            self.error("no connection to ES (not configured) skip ES indexing")
            return
        serializable = entity.cw_adapt_to(self.adapter)
        json = serializable.serialize()
        if not json:
            return
        es_cnx.index(
            index=self.index_name,
            id=serializable.es_id,
            doc_type=serializable.es_doc_type,
            body=json,
            params=params,
        )

    def es_delete(self, entity):
        es_cnx = self.get_connection()
        if es_cnx is None or not self.index_name:
            self.error("no connection to ES (not configured) skip ES deletion")
            return
        serializable = entity.cw_adapt_to(self.adapter)
        es_cnx.delete(
            index=self.index_name,
            id=serializable.es_id,
            doc_type=serializable.es_doc_type,
        )


class IFullTextIndexSerializable(view.EntityAdapter):
    """Adapter to serialize an entity to a bare python structure that may be
    directly serialized to e.g. JSON.
    """

    __regid__ = "IFullTextIndexSerializable"
    __select__ = is_instance("Any")
    custom_indexable_attributes = ()
    skip_indexable_attributes = ()

    @property
    def es_id(self):
        return self.entity.eid

    @property
    def es_doc_type(self):
        return "_doc"

    @cachedproperty
    def fulltext_indexable_attributes(self):
        eschema = self._cw.vreg.schema[self.entity.cw_etype]
        attrs = ["creation_date", "modification_date", "cwuri"]
        attrs.extend(
            [
                r.type
                for r in eschema.indexable_attributes()
                if r.type not in self.skip_indexable_attributes
            ]
        )
        for rschema, tschema in eschema.attribute_definitions():
            if rschema.type == "eid":
                continue
            # XXX
            if tschema.type in ("Int", "Float"):
                attrs.append(rschema.type)
        attrs.extend(self.custom_indexable_attributes)
        return attrs

    def process_attributes(self):
        data = {}
        for attr in self.fulltext_indexable_attributes:
            data[attr] = getattr(self.entity, attr)
        return data

    def serialize(self, complete=True):
        entity = self.entity
        if complete:
            entity.complete()
        data = {
            "cw_etype": entity.cw_etype,
            "eid": entity.eid,
            "cwuri": entity.cwuri,
        }
        # TODO take a look at what's in entity.cw_relation_cache
        data.update(self.process_attributes())
        return data


class File(IFullTextIndexSerializable):
    __select__ = IFullTextIndexSerializable.__select__ & is_instance("File")

    def serialize(self, complete=True):
        """this could be a generic implementation of fulltext_containers indexation, but for

        now we can not return more than one parent json which is fine
        for Files
        """
        for rschema, role in self._cw.vreg.schema["File"].fulltext_containers():
            for parent in self.entity.related(
                rschema.type, role=neg_role(role)
            ).entities():
                return parent.cw_adapt_to("IFullTextIndexSerializable").serialize(
                    complete
                )
        return {}


class ESTransactionQueue(EsRegistry):
    __regid__ = "es.opqueue"

    @cachedproperty
    def default_indexer(self):
        return self._cw.vreg["es"].select("indexer", self._cw)

    def purge_useless_operations(self, es_operations):
        """remove operations from `es_operations` that will have no effect.

        For instance:
        - if an entity is indexed several times, just index it once.
        - if an entity is indexed then deleted, simply remove the
          *index* operation.
        Since there are only two kind of operations ('delete' and 'index'),
        the general rule is that the last operation wins and therefore the
        first one is useless. The only corner case is when a delete operation
        is triggered before an index one. In that specific case, emit a warning
        and keep delete operation only.
        """
        done = collections.OrderedDict()
        for es_operation in reversed(es_operations):
            entity = es_operation["entity"]
            op_type = es_operation["op_type"]
            if entity.eid not in done:
                done[entity.eid] = es_operation
            else:
                prev_op_type = done[entity.eid]["op_type"]
                if op_type == "delete" and prev_op_type == "index":
                    self.warning(
                        "a delete operation on %s#%s inserted before" "an index one",
                        entity.cw_etype,
                        entity.eid,
                    )
                    done[entity.eid] = es_operation
        return done.values()

    def process_operation(self, es_operation):
        indexer = es_operation.get("indexer", self.default_indexer)
        entity = es_operation["entity"]
        if self._cw.deleted_in_transaction(entity.eid):
            es_method = indexer.es_delete
        elif es_operation["op_type"] == "index":
            es_method = partial(indexer.es_index, params={"refresh": "true"})
        elif es_operation["op_type"] == "delete":
            es_method = indexer.es_delete
        else:
            self.info(
                "skipping unknown operation type %s on %s",
                es_operation["op_type"],
                entity.eid,
            )
            return
        try:
            es_method(entity)
        except (ConnectionError, ProtocolError, NotFoundError) as exc:
            self.warning(
                "[ES] Failed to %s %s#%s (%s)",
                es_operation["op_type"],
                entity.cw_etype,
                entity.eid,
                exc,
            )

    def process_operations(self, es_operations):
        es_operations = self.purge_useless_operations(es_operations)
        for es_operation in es_operations:
            self.process_operation(es_operation)
        return es_operations



options = (
    (
        "elasticsearch-locations",
        {
            "type": "string",
            "default": "",
            "help": "Elastic Search location (eg. 192.168.0.23:9200), "
            "this can be a list of locations (192.168.0.23:9200,192.168.0.24:9200, "
            "you can also include the scheme (eg. http://192.168.0.23:9200) "
            "warning: if this is not defined indexing will be disabled (no localhost default)",
            "group": "elasticsearch",
            "level": 5,
        },
    ),
    (
        "index-name",
        {
            "type": "string",
            "default": "",
            "help": "Elastic Search index name (eg. cubicweb)"
            "warning: if this is not defined indexing will be disabled (no index name default)",
            "group": "elasticsearch",
            "level": 5,
        },
    ),
    (
        "elasticsearch-verify-certs",
        {
            "type": "yn",
            "default": True,
            "help": "Elastic Search connection verify_certs variable"
            "(set to false when using https with unknown or self-signed certificates)",
            "group": "elasticsearch",
            "level": 5,
        },
    ),
    (
        "elasticsearch-ssl-show-warn",
        {
            "type": "yn",
            "default": True,
            "help": "Elastic Search connection ssl-show-warn variable"
            "(set to no if you want to suppress ssl warnings when verify cert is no",
            "group": "elasticsearch",
            "level": 5,
        },
    ),
)



# -*- coding: utf-8 -*-
# copyright 2016-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import logging

from elasticsearch.exceptions import ConnectionError
from urllib3.exceptions import ProtocolError
from elasticsearch_dsl.connections import connections

from rql import parse as rql_parse
from rql.utils import rqlvar_maker

INDEXABLE_TYPES = None

# customization mechanism, in your cube, add your type as a key, and a list of
# additionnal attributes

log = logging.getLogger(__name__)


def indexable_types(schema, custom_skip_list=None):
    """
    introspect indexable types
    """
    global INDEXABLE_TYPES
    if INDEXABLE_TYPES is not None:
        return INDEXABLE_TYPES
    indexable_types = []
    skip_list = ["TrInfo", "EmailAddress"]
    if custom_skip_list:
        skip_list = skip_list + custom_skip_list
    for eschema in schema.entities():
        if eschema.type in skip_list:
            continue
        if not eschema.final:
            # check eschema.fulltext_relations() ? (skip wf_info_for ?
            # )
            if list(eschema.indexable_attributes()):
                indexable_types.append(eschema.type)
    INDEXABLE_TYPES = indexable_types
    return indexable_types


def fulltext_indexable_rql(etype, cnx, eid=None):
    """
    Generate RQL with fulltext_indexable attributes for a given entity type

    :eid:
       defaults to None, set it to an eid to get RQL for a single element (used in hooks)
    """
    varmaker = rqlvar_maker()
    V = next(varmaker)
    rql = ["WHERE %s is %s" % (V, etype)]
    if eid:
        rql.append("%s eid %i" % (V, eid))
    var = next(varmaker)
    selected = []
    cw_entity = cnx.vreg["etypes"].etype_class(etype)(cnx)
    for attr in cw_entity.cw_adapt_to(
        "IFullTextIndexSerializable"
    ).fulltext_indexable_attributes:
        var = next(varmaker)
        rql.append("%s %s %s" % (V, attr, var))
        selected.append(var)
    return "Any %s,%s %s" % (V, ",".join(selected), ",".join(rql))


def indexable_entities(cnx, etype, chunksize=100000):
    """yield indexable entities of type `etype`

    This function uses `fulltext_indexable_rql` to build the appropriate RQL
    query for `etype` and then adds ``ORDERBY`` / ``LIMIT`` clauses to build
    rsets of at most `chunksize` length.

    Instead of doing::

        rql = fulltext_indexable_rql(etype, cnx)
        rset = cnx.execute(rql)
        for entity in rset.entities():
            # do something with entity

    You can therefore use this drop-in "memory efficient" alternative:

        for entity in indexable_entities(cnx, etype):
            # do something with entity
    """
    rql = fulltext_indexable_rql(etype, cnx)
    rqlst = rql_parse(rql).children[0]
    rqlst.set_limit(chunksize)
    mainvar = next(rqlst.get_selected_variables())
    rqlst.add_sort_var(mainvar)
    last_eid = 0
    while True:
        rqlst.save_state()
        rqlst.add_constant_restriction(mainvar, "eid", last_eid, "Int", ">")
        rql = rqlst.as_string()
        rqlst.recover()
        cnx.debug(u"RQL: {}".format(rql))
        rset = cnx.execute(rql)
        if not rset:
            break
        for entity in rset.entities():
            yield entity
            entity.cw_clear_all_caches()
        cnx.drop_entity_cache()
        last_eid = rset[-1][0]


def create_index(es, index_name, settings=None):
    """Create ``index_name`` if it doesn't already exist in ES


    Parameters
    ----------

    :es:
      the elastic search connection

    :index_name:
      the index name

    :settings:
      mapping and analyzer definitions

    """
    try:
        if index_name and not es.indices.exists(index=index_name):
            es.indices.create(index=index_name, body=settings)
    except (ConnectionError, ProtocolError):
        log.debug("Failed to index in hook, could not connect to ES")


def get_connection(config):
    """
    Get connection with config object, creates a persistent connexion and
    """
    try:
        return connections.get_connection()
    except KeyError:
        locations = config["elasticsearch-locations"]
        if locations:
            # TODO sanitize locations
            es = connections.create_connection(
                hosts=locations.split(","),
                verify_certs=config["elasticsearch-verify-certs"],
                ssl_show_warn=config["elasticsearch-ssl-show-warn"],
                timeout=20,
            )
            return es
        # TODO else ? raise KeyError - return None is OK?



# -*- coding: utf-8 -*-
"""cubicweb-ctl plugin providing the index-in-es command

:organization: Logilab
:copyright: 2016-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
from __future__ import print_function

import os.path as osp

from elasticsearch.helpers import parallel_bulk

from cubicweb.cwctl import CWCTL, init_cmdline_log_threshold
from cubicweb.cwconfig import CubicWebConfiguration as cwcfg
from cubicweb.toolsutils import Command

from cubicweb_elasticsearch.es import indexable_types, indexable_entities

HERE = osp.dirname(osp.abspath(__file__))


class IndexInES(Command):
    """Index content in ElasticSearch.

    <instance id>
      identifier of the instance

    """

    name = "index-in-es"
    min_args = max_args = 1
    arguments = "<instance id>"
    options = [
        (
            "dry-run",
            {
                "action": "store_true",
                "default": False,
                "short": "N",
                "help": "set to True if you want to skip the insertion in ES",
            },
        ),
        (
            "debug",
            {
                "action": "store_true",
                "default": False,
                "short": "D",
                "help": ("shortcut for --loglevel=debug"),
            },
        ),
        (
            "loglevel",
            {
                "short": "l",
                "type": "choice",
                "metavar": "<log level>",
                "default": None,
                "choices": ("debug", "info", "warning", "error"),
            },
        ),
        (
            "etypes",
            {
                "type": "csv",
                "default": "",
                "help": "only index given etypes [default:all indexable types]",
            },
        ),
        (
            "index-name",
            {
                "type": "string",
                "default": "",
                "help": (
                    "override index-name if you want to use a different ID"
                    "[default: uses index-name from all-in-one.conf]"
                ),
            },
        ),
        (
            "except-etypes",
            {
                "type": "string",
                "default": "",
                "help": "all indexable types except given etypes" "[default: []]",
            },
        ),
        (
            "chunksize",
            {
                "type": "int",
                "default": 100000,
                "help": "max number of entities to fetch at once (deafult: 100000)",
            },
        ),
    ]

    def run(self, args):
        """run the command with its specific arguments"""
        appid = args.pop(0)
        if self["debug"]:
            self["loglevel"] = "debug"
        config = cwcfg.config_for(appid, debugmode=self["loglevel"])
        if self["loglevel"]:
            init_cmdline_log_threshold(config, self["loglevel"])
        with config.repository().internal_cnx() as cnx:
            schema = cnx.vreg.schema
            indexer = cnx.vreg["es"].select("indexer", cnx)
            es = indexer.get_connection()
            indexer.create_index()
            if self.config.index_name:
                cnx.info("create ES index {}".format(self.config.index_name))
                indexer.create_index(index_name=self.config.index_name)
            if es:
                if self.config.etypes:
                    etypes = self.config.etypes
                else:
                    etypes = indexable_types(
                        schema, custom_skip_list=self.config.except_etypes
                    )
                    assert self.config.except_etypes not in etypes
                if not self.config.etypes:
                    cnx.debug(u"found indexable types: {}".format(",".join(etypes)))
                for _ in parallel_bulk(
                    es,
                    self.bulk_actions(
                        etypes,
                        cnx,
                        index_name=self.config.index_name,
                        dry_run=self.config.dry_run,
                    ),
                    raise_on_error=False,
                    raise_on_exception=False,
                ):
                    pass
            else:
                cnx.info(u"no elasticsearch configuration found, skipping")

    def bulk_actions(self, etypes, cnx, index_name=None, dry_run=False):
        if index_name is None:
            index_name = cnx.vreg.config["index-name"]
        for etype in etypes:
            for idx, entity in enumerate(
                indexable_entities(cnx, etype, chunksize=self.config.chunksize), 1
            ):
                try:
                    serializer = entity.cw_adapt_to("IFullTextIndexSerializable")
                    json = serializer.serialize(complete=False)
                except Exception:
                    cnx.error(
                        "[{}] Failed to serialize entity {} ({})".format(
                            index_name, entity.eid, etype
                        )
                    )
                    continue
                if not dry_run and json:
                    # Entities with
                    # fulltext_containers relations return their container
                    # IFullTextIndex serializer , therefor the "id" and
                    # "doc_type" in kwargs bellow must be container data.
                    data = {
                        "_op_type": "index",
                        "_index": index_name or cnx.vreg.config["index-name"],
                        "_id": serializer.es_id,
                        "_source": json,
                    }
                    self.customize_data(data)
                    yield data
            cnx.info(u"[{}] indexed {} {} entities".format(index_name, idx, etype))

    def customize_data(self, data):
        pass


CWCTL.register(IndexInES)



# -*- coding: utf-8 -*-
# copyright 2016-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""elasticsearch search views"""
from six import text_type as unicode

from elasticsearch.exceptions import NotFoundError
from elasticsearch_dsl import FacetedSearch, TermsFacet, DateHistogramFacet
from bs4 import BeautifulSoup

from logilab.mtconverter import xml_escape

import cwtags.tag as t
from cubicweb import _
from cubicweb.view import StartupView

from cubicweb_elasticsearch.es import get_connection
from cubicweb_elasticsearch.search_helpers import compose_search


def normalize_value(value):
    # FIXME TODO have better typing mechanisme (inspect facets?)
    try:
        if value.lower() == "true":
            return True
        elif value.lower() == "false":
            return False
        else:
            return int(value)
    except ValueError:
        pass
    return value


class CWFacetedSearch(FacetedSearch):
    # fields that should be searched
    fields = ["title^3", "description^2", "alltext"]

    facets = {
        # use bucket aggregations to define facets
        "cw_etype": TermsFacet(field="cw_etype"),
        "creation_date": DateHistogramFacet(field="creation_date", interval="month"),
    }

    def __init__(
        self,
        query=None,
        filters={},
        sort=(),
        doc_types=None,
        index=None,
        form=None,
        track_total_hits=True,
        **kwargs
    ):
        if index:
            self.index = index
        if doc_types:
            self.doc_types = doc_types
        if form:
            self.form = form
        else:
            self.form = {}
        # Count all the hits by default
        self.track_total_hits = track_total_hits
        self.extra_kwargs = kwargs
        super(CWFacetedSearch, self).__init__(query, filters, sort=sort)

    def search(self):
        # override methods to add custom pieces
        s = super(CWFacetedSearch, self).search()
        return s.extra(track_total_hits=self.track_total_hits)

    def query(self, search, query):
        if query:
            common = "debug-es-disable-common" not in self.form  # default True
            phrase = "debug-es-disable-phrase" not in self.form  # default True
            fuzzy = "fuzzy" in self.form  # default False
            return compose_search(
                search,
                query=query,
                fields=self.fields,
                fuzzy=fuzzy,
                common=common,
                phrase=phrase,
            )
        return search

    def highlight(self, search):
        """
        Add custom highlighting
        """
        search = search.highlight(
            *(f if "^" not in f else f.split("^", 1)[0] for f in self.fields)
        )
        return search.highlight_options(fragment_size=150, encoder="html")


class ElasticSearchView(StartupView):
    __regid__ = "esearch"
    previous_link = u" &lt; "
    next_link = u" &gt; "
    middle_link = u" &middot; " * 3
    title = _("Search")

    def render_search_comp(self):
        search_comp = self._cw.vreg["components"].select_or_none(
            "search-comp", self._cw
        )
        if search_comp:
            search_comp.render(w=self.w)

    def do_search(self, query_string):
        get_connection(self._cw.vreg.config)
        facet_selections = {}
        items_per_page = int(self._cw.form.get("items_per_page", 10))
        start, stop = 0, items_per_page
        for key, value in self._cw.form.items():
            if key.startswith("es_"):
                if isinstance(value, list):
                    for index in range(len(value)):
                        value[index] = normalize_value(value[index])
                else:
                    value = normalize_value(value)

                facet_selections[key.replace("es_", "")] = value
            if key == "page":
                try:
                    start = (max(int(value) - 1, 0)) * items_per_page
                    stop = start + items_per_page
                except ValueError:
                    pass
        search = self.customize_search(query_string, facet_selections, start, stop)
        if "debug-es" in self._cw.form:
            import json

            self.w(u"<pre>")
            self.w(unicode(json.dumps(search._s.to_dict(), indent=2)))
            self.w(u"</pre>")
            self.w(u"<br/>")
        try:
            response = search.execute()
            if "debug-es" in self._cw.form:
                import json

                self.w(u"<pre>")
                self.w(unicode(json.dumps(response.to_dict(), indent=2)))
                self.w(u"</pre>")
                self.w(u"<br/>")
            return response
        except NotFoundError:
            self.w(u"index not found in elasticsearch")
            return

    def call(self, **kwargs):
        # TODO if no ES configuration, redirect or display warning
        self.render_search_comp()
        query_string = self._cw.form.get("q", self._cw.form.get("search", ""))
        self.w(u"<h1>%s</h1>" % self._cw._(self.title))
        response = self.do_search(query_string)
        if response.hits.total.value:
            self.w(u"<h2>Resultats pour : <em>%s</em></h2>" % xml_escape(query_string))
            self.w(u"Resultats: %s" % response.hits.total.value)
            if hasattr(response, "facets"):
                self.display_facets(response)
        self.display_results(response)

    def customize_search(
        self, query_string, facet_selections, start=0, stop=10, **kwargs
    ):
        """
        This is where one can customize the search by modifying the
        query string and facet selection in an inherited class.

        For example :
        * add specific keywords sur as id:text and
          add them the facet_selection
        * use your own CWFacetedSearch class to modify fields
          and facets
        """
        return CWFacetedSearch(
            query_string,
            facet_selections,
            index=self._cw.vreg.config.get("index-name"),
            doc_types=["_doc"],
            form=self._cw.form,
            **kwargs
        )[start:stop]

    def display_results(self, response):
        """
        Display results obtained from elasticsearch
        """
        self.w(u'<div id="main-center" class="col-xs-10" role="main">')
        self.pagination(response)
        self.w(u"<ul>")
        for result in response:
            self.w(u"<li>")
            infos = result.to_dict()
            infos["_score"] = result.meta.score
            infos["keys"] = result.to_dict().keys()
            infos["url"] = (
                infos["cwuri"].startswith("_auto_generated")
                and infos["eid"]
                or infos["cwuri"]
            )
            self.customize_infos(infos)
            try:
                self.w(u'<a href="%(url)s">%(title)s</a> (%(_score).2f)<br/>' % (infos))
                if self._cw.form.get("debug-es"):
                    self.w(u" [%(keys)s] <br/>" % infos)
            except KeyError:
                self.w(u"Missing key in : %s" % infos.keys())
            try:
                for fragment in result.meta.highlight.content:
                    self.w(u"... %s" % BeautifulSoup(fragment, "lxml").get_text())
                    self.w(u" ... <br/>")
            except AttributeError:
                pass
            self.w(u"</li>")
        self.w(u"</ul>")
        self.pagination(response)
        self.w(u"</div>")

    def customize_infos(self, infos):
        """
        This is where one can customize the infos being displayed

        For example : set the title according to your rules and data set
        """
        pass

    def pagination(self, response):
        """
        Pagination HTML generation
        """
        if response.hits.total.value <= 10:
            return
        url_params = self._cw.form.copy()
        with t.ul(self.w, klass="pagination") as ul:
            current_page = int(url_params.get("page", 1))
            url_params["page"] = current_page - 1
            if current_page - 1 >= 1:
                ul(
                    t.li(
                        t.a(
                            self.previous_link,
                            href=xml_escape(self._cw.build_url(**url_params)),
                        )
                    )
                )
            else:
                ul(t.li(t.a(self.previous_link)))
            total_pages = min((response.hits.total.value // 10) + 2, 1000)
            page_padding = 3

            if current_page > page_padding:
                for page in range(
                    1, min(page_padding + 1, current_page - page_padding)
                ):
                    self.page_number(url_params, page, current_page, ul)
                if current_page > (page_padding * 2) + 1:
                    ul(t.li(t.a(self.middle_link)))
            for page in range(
                max(1, current_page - page_padding),
                min(current_page + page_padding, total_pages),
            ):
                self.page_number(url_params, page, current_page, ul)
            if current_page < total_pages - page_padding:
                if current_page < total_pages - page_padding * 2:
                    ul(t.li(t.a(self.middle_link)))
                for page in range(
                    max(current_page + page_padding, total_pages - page_padding),
                    total_pages,
                ):
                    self.page_number(url_params, page, current_page, ul)

            url_params["page"] = current_page + 1
            if current_page + 1 >= (total_pages):
                ul(t.li(t.a(self.next_link)))
            else:
                ul(
                    t.li(
                        t.a(
                            self.next_link,
                            href=xml_escape(self._cw.build_url(**url_params)),
                        )
                    )
                )

    def page_number(self, url_params, page, current_page, ul):
        """
        Generate HTML for page number (bold if page is current_page)
        """
        url_params["page"] = page
        url = self._cw.build_url(**url_params)
        if page == current_page:
            ul(
                t.li(
                    t.a(t.b(page), href=xml_escape(url)),
                    Class="active",
                )
            )
        else:
            ul(t.li(t.a(page, href=xml_escape(url))))
        return url

    @property
    def facets_to_display(self):
        """
        Method to list facets to display (can be customized)
        """
        return ("cw_etype",)

    def display_facets(self, response):
        """
        Generate HTML for facets
        """
        self.w(
            u"""<aside id="aside-main-left" class="col-xs-2 cwjs-aside">
                   <div class="panel panel-default contextFreeBox facet_filterbox">
                      <div class="panel-heading">
                         <div class="panel-title">Facettes</div>
                      </div>
        """
        )
        for attribute in self.facets_to_display:
            url_params = self._cw.form.copy()
            if "page" in url_params:
                del url_params["page"]
            self.w(u'<div class="facetBody vocabularyFacet">')
            self.w(u'<div class="facetTitle">{}</div>'.format(attribute))
            for (tag, count, selected) in response.facets[attribute]:
                # facetValueSelected / facetValueDisabled in class
                facet_item = (
                    u'<div class="facetValue facetCheckBox">'
                    "    <span>"
                    '      <a href="{}">{} {}</a>'
                    "    </span>"
                    "</div>"
                )
                if url_params.get("es_{}".format(attribute)) != tag:
                    url_params["es_{}".format(attribute)] = str(tag)
                else:
                    del url_params["es_{}".format(attribute)]
                url = self._cw.build_url(**url_params)
                content = (
                    selected and '<div class="facet-active">{}</div>'.format(tag) or tag
                )
                self.w(facet_item.format(url, content, count))
            self.w(u"</div>")
        self.w(u"</div></aside>")



# -*- coding: utf-8 -*-
# copyright 2016-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from elasticsearch_dsl import Q, query as dsl_query

from logilab.mtconverter import xml_escape


def compose_search(
    search, query=None, fields=(), fuzzy=False, phrase=True, common=True
):
    """
    Compose a elasticsearch-dsl query from queries :

    * simple term
    * simple terms (OR)
    * negation (add - in front of a term)
    * explicit OR
    * quoted terms (AND)

    search:
        search object (used to set doc_type and index_name outside of
        compose_search)
    query:
        text of the query to be composed (can contain quotes)
    fields:
        restrict and boost search on certain fields eg. ('title^2', 'alltext')
    fuzzy:
        add a fuzzy search element to part of the query generated
        https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-fuzzy-query.html
    """
    # FIXME TODO - restructure entier code base, have a proper lexer
    for char in ('"', "'", xml_escape('"'), xml_escape("'")):
        # TODO - implement phrase + term
        if len(query.split(char)) == 3:
            # TODO add this to most important queries, instead of single query ?
            return search.query(
                "bool",
                must=Q(
                    "multi_match",
                    query=query.split(char)[1],
                    type="phrase",
                    fields=fields,
                ),
            )
    must = []
    must_not = []
    should = []
    cutoff_frequency = 0.001
    # https://www.elastic.co/guide/en/elasticsearch/reference/2.4/query-dsl-minimum-should-match.html
    minimum_should_match = "1"
    # proximity booster - phrase with slop=50

    # TODO find a better way to do this
    if not phrase and not common:  # invalid combination
        phrase = common = True
    if phrase:
        phrase_query = Q(
            "multi_match", query=query, type="phrase", slop=50, fields=fields
        )
        should.append(phrase_query)
    # highfrequency/lowfrequency query
    # https://www.elastic.co/blog/stop-stopping-stop-words-a-look-at-common-terms-query
    if common:
        common_query = dsl_query.Common(
            alltext={
                "query": query,
                "cutoff_frequency": cutoff_frequency,
                "low_freq_operator": "and",
                "minimum_should_match": {"high_freq": "70%"},
            }
        )
        should.append(common_query)

    elements = query.split()
    elements_lowercase = [e.lower() for e in elements]
    if "or" in elements_lowercase and len(elements) >= 3:
        for element in query.split("or"):
            should.append(Q("multi_match", query=element.strip(), fields=fields))
        elements = []
    for element in elements:
        if element.startswith("-"):
            must_not.append(Q("multi_match", query=element[1:], fields=fields))

        else:
            if fuzzy:
                should.append(dsl_query.Fuzzy(alltext=element))
    bool_query = dsl_query.Bool(
        must=must,
        must_not=must_not,
        should=should,
        minimum_should_match=minimum_should_match,
    )
    search.query = bool_query
    return search



import unittest

from logilab.common.registry import yes

from cubicweb.devtools import testlib

from cubicweb_elasticsearch.entities import ESTransactionQueue


class MockQueue(ESTransactionQueue):
    __select__ = yes(10)
    es_operations = []

    def process_operation(self, es_operation):
        self.es_operations.append(
            (
                es_operation["op_type"],
                es_operation["entity"].cw_etype,
                es_operation["entity"].eid,
            )
        )

    def process_operations(self, es_operations):
        MockQueue.es_operations = []
        super(MockQueue, self).process_operations(es_operations)


class IndexHookTC(testlib.CubicWebTC):
    def test_esqueue_index(self):
        """ensure queue gets filled with "index" operations"""
        with self.admin_access.cnx() as cnx:
            with self.temporary_appobjects(MockQueue):
                ce = cnx.create_entity
                p = ce("Person", age=12, name=u"Jean")
                cnx.commit()
            self.assertEqual(
                MockQueue.es_operations,
                [
                    ("index", "Person", p.eid),
                ],
            )

    def test_esqueue_index_multiple(self):
        """ensure queue gets filled with "index" operations"""
        with self.admin_access.cnx() as cnx:
            with self.temporary_appobjects(MockQueue):
                ce = cnx.create_entity
                p = ce("Person", age=12, name=u"Jean")
                p2 = ce("Person", age=13, name=u"Jeanne")
                cnx.commit()
            self.assertCountEqual(
                MockQueue.es_operations,
                [
                    ("index", "Person", p.eid),
                    ("index", "Person", p2.eid),
                ],
            )

    def test_esqueue_delete(self):
        """ensure queue gets filled with "index" operations"""
        with self.temporary_appobjects(MockQueue):
            with self.admin_access.cnx() as cnx:
                ce = cnx.create_entity
                p = ce("Person", age=12, name=u"Jean")
                p2 = ce("Person", age=13, name=u"Jeanne")
                cnx.commit()
            with self.admin_access.cnx() as cnx:
                cnx.execute("DELETE Person P")
                cnx.commit()
            self.assertCountEqual(
                MockQueue.es_operations,
                [
                    ("delete", "Person", p.eid),
                    ("delete", "Person", p2.eid),
                ],
            )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# copyright 2016-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-elasticsearch specific hooks and operations"""

from cubicweb.server import hook
from cubicweb.predicates import score_entity

from cubicweb_elasticsearch.es import indexable_types


def entity_indexable(entity):
    """ returns boolean to indicate if entity is indexable """
    return (
        entity.cw_etype in indexable_types(entity._cw.vreg.schema)
        or entity.cw_adapt_to("IFullTextIndexSerializable").custom_indexable_attributes
    )


class ContentUpdateIndexES(hook.Hook):
    """detect content change and updates ES indexing"""

    __regid__ = "elasticsearch.contentupdatetoes"
    __select__ = hook.Hook.__select__ & score_entity(entity_indexable)
    events = ("after_update_entity", "after_add_entity", "after_delete_entity")
    category = "es"

    def __call__(self):
        op_type = "delete" if self.event == "after_delete_entity" else "index"
        IndexEsOperation.get_instance(self._cw).add_data(
            {
                "op_type": op_type,
                "entity": self.entity,
            }
        )


class RelationsUpdateIndexES(hook.Hook):
    """detect relations changes and updates ES indexing"""

    __regid__ = "elasticsearch.relationsupdatetoes"
    events = ("after_add_relation", "before_delete_relation")
    category = "es"

    def __call__(self):
        # XXX add a selector for object and subject
        for entity in (
            self._cw.entity_from_eid(self.eidfrom),
            self._cw.entity_from_eid(self.eidto),
        ):
            if entity_indexable(entity):
                IndexEsOperation.get_instance(self._cw).add_data(
                    {
                        "op_type": "index",
                        "entity": entity,
                    }
                )


class IndexEsOperation(hook.DataOperationMixIn, hook.Operation):
    """ mixin class to process ES indexing as a postcommit event """

    containercls = list

    def postcommit_event(self):
        queue = self.cnx.vreg["es"].select("es.opqueue", req=self.cnx)
        queue.process_operations(self.get_data())



import unittest

from mock import patch

from cubicweb.devtools import testlib
from cubicweb.cwconfig import CubicWebConfiguration


class IndexHookTC(testlib.CubicWebTC):
    def setup_database(self):
        super(IndexHookTC, self).setup_database()
        self.orig_config_for = CubicWebConfiguration.config_for

        def config_for(appid):
            return self.config  # noqa

        CubicWebConfiguration.config_for = staticmethod(config_for)
        self.config[
            "elasticsearch-locations"
        ] = "http://nonexistant.elastic.search:9200"
        self.config["index-name"] = "unittest_index_name"

    @patch("elasticsearch.client.indices.IndicesClient.create")
    @patch("elasticsearch.client.indices.IndicesClient.exists")
    @patch("elasticsearch.client.Elasticsearch.index")
    def test_index_after_create_entity(self, create, exists, index):
        with self.admin_access.cnx() as cnx:
            indexer = cnx.vreg["es"].select("indexer", cnx)
            ce = cnx.create_entity
            p = ce("Person", age=12, name=u"Jean")
            cnx.commit()
            es = indexer.get_connection()
            self.assertTrue(es.index.called)
            args, kwargs = es.index.call_args
            for arg_name, expected_value in (("id", p.eid), ("doc_type", "_doc")):
                self.assertEqual(kwargs[arg_name], expected_value)


if __name__ == "__main__":
    unittest.main()



import unittest

try:
    from http.client import HTTPConnection
except ImportError:
    from httplib import HTTPConnection

from elasticsearch_dsl.connections import connections

from cubicweb.predicates import is_instance

from cubicweb_elasticsearch.entities import IFullTextIndexSerializable


class BlogEntryFTIAdapter(IFullTextIndexSerializable):
    __select__ = IFullTextIndexSerializable.__select__ & is_instance("BlogEntry")
    custom_indexable_attributes = ("title", "content")


class BlogFTIAdapter(IFullTextIndexSerializable):
    __select__ = IFullTextIndexSerializable.__select__ & is_instance("Blog")
    custom_indexable_attributes = ("title",)


class PersonFTIAdapter(IFullTextIndexSerializable):
    __select__ = IFullTextIndexSerializable.__select__ & is_instance("Person")

    @property
    def es_id(self):
        return self.entity.age


class RealESTestMixin(object):
    @classmethod
    def setUpClass(cls):
        try:
            HTTPConnection("localhost:9200").request("GET", "/")
        except:  # noqa
            raise unittest.SkipTest("No ElasticSearch on localhost, skipping test")
        super(RealESTestMixin, cls).setUpClass()

    def setup_database(self):
        super(RealESTestMixin, self).setup_database()
        self.config.global_set_option(
            "elasticsearch-locations", "http://localhost:9200"
        )
        self.config.global_set_option("index-name", "unittest_index_name")

    def tearDown(self):
        try:
            with self.admin_access.cnx() as cnx:
                indexer = cnx.vreg["es"].select("indexer", cnx)
                es = indexer.get_connection()
                es.indices.delete(self.config["index-name"])
        finally:
            # remove default connection if there's one
            try:
                connections.remove_connection("default")
            except KeyError:
                pass
            super(RealESTestMixin, self).tearDown()



# copyright 2003-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
import os
import sys

from logilab.common.pytest import PyTester


def getlogin():
    """avoid usinng os.getlogin() because of strange tty / stdin problems
    (man 3 getlogin)
    Another solution would be to use $LOGNAME, $USER or $USERNAME
    """
    if sys.platform == "win32":
        return os.environ.get("USERNAME") or "cubicweb"
    import pwd

    return pwd.getpwuid(os.getuid())[0]


def update_parser(parser):
    login = getlogin()
    parser.add_option(
        "-r",
        "--rebuild-database",
        dest="rebuild_db",
        default=False,
        action="store_true",
        help="remove tmpdb and rebuilds the test database",
    )
    parser.add_option(
        "-u",
        "--dbuser",
        dest="dbuser",
        action="store",
        default=login,
        help="database user",
    )
    parser.add_option(
        "-w",
        "--dbpassword",
        dest="dbpassword",
        action="store",
        default=login,
        help="database user's password",
    )
    parser.add_option(
        "-n",
        "--dbname",
        dest="dbname",
        action="store",
        default=None,
        help="database name",
    )
    parser.add_option(
        "--euser", dest="euser", action="store", default=login, help="euser name"
    )
    parser.add_option(
        "--epassword",
        dest="epassword",
        action="store",
        default=login,
        help="euser's password' name",
    )
    return parser


class CustomPyTester(PyTester):
    def __init__(self, cvg, options):
        super(CustomPyTester, self).__init__(cvg, options)
        if options.rebuild_db:
            os.unlink("tmpdb")
            os.unlink("tmpdb-template")



import unittest

from mock import patch

from cubicweb.devtools import testlib
from cubicweb.cwconfig import CubicWebConfiguration

from cubicweb_elasticsearch.testutils import (
    BlogFTIAdapter,
    BlogEntryFTIAdapter,
    PersonFTIAdapter,
)


class IFullTextIndexSerializableTC(testlib.CubicWebTC):
    def setup_database(self):
        super(IFullTextIndexSerializableTC, self).setup_database()
        self.orig_config_for = CubicWebConfiguration.config_for

        def config_for(appid):
            return self.config  # noqa

        CubicWebConfiguration.config_for = staticmethod(config_for)
        self.config[
            "elasticsearch-locations"
        ] = "http://nonexistant.elastic.search:9200"
        self.config["index-name"] = "unittest_index_name"

    @patch("elasticsearch.client.indices.IndicesClient.create")
    @patch("elasticsearch.client.indices.IndicesClient.exists")
    @patch("elasticsearch.client.Elasticsearch.index")
    def test_index_entity(self, index, exists, create):
        """Only update indexable attributes while call entity.complete()
        on IFullTextIndexSerializable.serialze()
        """
        with self.admin_access.repo_cnx() as cnx:
            with self.temporary_appobjects(BlogFTIAdapter, BlogEntryFTIAdapter):
                blog = cnx.create_entity("Blog", title=u"Blog")
                cnx.commit()
                self.assertTrue(index.called)
                args, kwargs = index.call_args
                # blog title is a in custom_indexable_attributes
                self.assertEqual(kwargs["doc_type"], "_doc")
                self.assertEqual(kwargs["body"]["cw_etype"], "Blog")
                self.assertEqual(kwargs["body"]["title"], u"Blog")
                index.reset_mock()
                # create a BlogEntry
                bentry = cnx.create_entity(
                    "BlogEntry",
                    title=u"program",
                    content=u"Le nouveau programme",
                    entry_of=blog,
                )
                cnx.commit()
                self.assertEqual(index.call_count, 2)
                for args, kwargs in index.call_args_list:
                    if kwargs["doc_type"] == "_doc":
                        break
                else:
                    self.fail("index not called for the BlogEntry")
                for arg_name, expected_value in (
                    ("content", u"Le nouveau programme"),
                    ("cwuri", bentry.cwuri),
                    ("title", "program"),
                ):
                    self.assertEqual(kwargs["body"][arg_name], expected_value)
                self.assertFalse("content_format" in kwargs["body"])
                # update BlogEntry
                bentry.cw_set(title=u"Programme")
                index.reset_mock()
                cnx.commit()
                self.assertTrue(index.called)
                args, kwargs = index.call_args
                for arg_name, expected_value in (
                    ("id", bentry.eid),
                    ("doc_type", "_doc"),
                ):
                    self.assertEqual(kwargs[arg_name], expected_value)
                for arg_name, expected_value in (
                    ("content", u"Le nouveau programme"),
                    ("cwuri", bentry.cwuri),
                    ("title", u"Programme"),
                ):
                    self.assertEqual(kwargs["body"][arg_name], expected_value)
                self.assertFalse("content_format" in kwargs["body"])

    @patch("elasticsearch.client.indices.IndicesClient.create")
    @patch("elasticsearch.client.indices.IndicesClient.exists")
    @patch("elasticsearch.client.Elasticsearch.index")
    def test_custom_es_id_attr(self, create, exists, index):
        """check custom es_eid_attr is used in es document"""
        with self.admin_access.repo_cnx() as cnx:
            with self.temporary_appobjects(PersonFTIAdapter):
                cnx.create_entity("Person", age=123456, name=u"Jean")
                cnx.commit()
                indexer = cnx.vreg["es"].select("indexer", cnx)
                es = indexer.get_connection()
                self.assertTrue(es.index.called)
                args, kwargs = es.index.call_args
                # make sure age was used as custom es document id
                self.assertEqual(kwargs["id"], 123456)


class SerializationTests(testlib.CubicWebTC):
    def test_person_serialization(self):
        with self.admin_access.cnx() as cnx:
            jean = cnx.create_entity("Person", age=12, name=u"Jean")
            serializer = jean.cw_adapt_to("IFullTextIndexSerializable")
            self.assertEqual(
                serializer.serialize(),
                {
                    "name": u"Jean",
                    "age": 12,
                    "cw_etype": u"Person",
                    "eid": jean.eid,
                    "cwuri": jean.cwuri,
                    "creation_date": jean.creation_date,
                    "modification_date": jean.modification_date,
                },
            )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=W0622
"""cubicweb-elasticsearch application packaging information"""

modname = "elasticsearch"
distname = "cubicweb-elasticsearch"

numversion = (0, 9, 2)
version = ".".join(str(num) for num in numversion)

license = "LGPL"
author = "LOGILAB S.A. (Paris, FRANCE)"
author_email = "contact@logilab.fr"
description = "Simple ElasticSearch indexing integration for CubicWeb"
web = "http://www.cubicweb.org/project/%s" % distname

__depends__ = {
    "cubicweb": ">= 3.24.0",
    "six": ">= 1.4.0",
    "cwtags": None,
    "elasticsearch": ">=7.0.0,<8.0.0",
    "elasticsearch-dsl": ">=7.0.0,<8.0.0",
    "beautifulsoup4": None,
}

__recommends__ = {}

classifiers = [
    "Environment :: Web Environment",
    "Framework :: CubicWeb",
    "Programming Language :: Python",
    "Programming Language :: JavaScript",
]
